/*
 * Title:        sag_internal/plugin_macros.hpp
 * Description:  Implementation of macros for generating C ABI for plugins
 * $Copyright (c) 2015-2020 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: plugin_macros.hpp 370738 2020-04-09 10:32:47Z matj $
 */

#ifndef _SAG_INTERNAL_PLUGIN_MACROS_HPP_
#define _SAG_INTERNAL_PLUGIN_MACROS_HPP_

/** @private
 * @file Exports the C functions implemented by the plugin. 
*/

////////////////////////////////////////////////////////////////////////////////////////////
// Shared between codecs and transports

/** @internal C functions for API common to all plug-ins; excludes constructor */ 
#define _SAG_DECLARE_CONNECTIVITY_CORE(Name, Class)	\
extern "C" {\
SAG_CONNECTIVITY_DLL_SYM uint64_t SAG_CONNECTIVITY_CALL sag_plugin_api_version_##Name() \
{\
	return SAG_PLUGIN_API_VERSION;\
}\
SAG_CONNECTIVITY_DLL_SYM const char* SAG_CONNECTIVITY_CALL sag_plugin_product_version_##Name() \
{\
	return SAG_PLUGIN_PRODUCT_VERSION;\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_plugin_start_##Name(sag_plugin_t plugin)\
{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->start();\
		return SAG_ERROR_OK;\
	} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::Plugin*>(plugin.plugin)->logger, SAG_LOG_ERROR, "start");\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_plugin_shutdown_##Name(sag_plugin_t plugin)\
{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->shutdown();\
		return SAG_ERROR_OK;\
	} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::Plugin*>(plugin.plugin)->logger, SAG_LOG_WARN, "shutdown");\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_destroy_plugin_##Name(sag_plugin_t plugin)\
{\
	try {\
		delete reinterpret_cast<Class*>(plugin.plugin); \
		return SAG_ERROR_OK;\
	} catch (...) {\
		/*shouldn't throw, and we couldn't log it if it did*/\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_plugin_hostReady_##Name(sag_plugin_t plugin)\
{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->hostReady();\
		return SAG_ERROR_OK;\
	} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::Plugin*>(plugin.plugin)->logger, SAG_LOG_WARN, "hostReady");\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_plugin_setNextTowardsHost_##Name(sag_plugin_t plugin, sag_plugin_t next, sag_send_fn_t fn)\
{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->setNextTowardsHost(com::softwareag::connectivity::HostSide::ptr_t(new com::softwareag::connectivity::RemoteHostSide(next, fn)));\
		return SAG_ERROR_OK;\
	} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::Plugin*>(plugin.plugin)->logger, SAG_LOG_ERROR, "setNextTowardsHost");\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_plugin_sendBatchTowardsTransport_##Name(sag_plugin_t plugin, sag_underlying_message_t *start, sag_underlying_message_t *end)\
{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->sendBatchTowardsTransport(reinterpret_cast<com::softwareag::connectivity::Message *>(start), reinterpret_cast<com::softwareag::connectivity::Message *>(end)); \
		return SAG_ERROR_OK;\
	} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::Plugin*>(plugin.plugin)->logger, SAG_LOG_WARN, "sendBatchTowardsTransport");\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
}

////////////////////////////////////////////////////////////////////////////////////////////
// Transports


/** @internal */
#define _SAG_DECLARE_CONNECTIVITY_TRANSPORT_COMMON(Name, Class)	\
_SAG_DECLARE_CONNECTIVITY_CORE(Name, Class)\
extern "C" {\
	void _sag_plugin_checkIsTransport_##Name(Class *clazz) {\
		com::softwareag::connectivity::AbstractTransport *ac = clazz;\
		(void) ac;\
	}\
}

/** @internal */
#define _SAG_DECLARE_CONNECTIVITY_TRANSPORT_CLASS(Class) \
_SAG_DECLARE_CONNECTIVITY_TRANSPORT_COMMON(Class, Class)\
extern "C" {\
SAG_CONNECTIVITY_DLL_SYM sag_plugin_t SAG_CONNECTIVITY_CALL sag_create_plugin_with_params_##Class(const char *name, const char *chainId, sag_underlying_data_t config, void* connectivityManager, void* reserved)\
	{\
		sag_plugin_t plugin;\
		plugin.plugin = 0;\
		try {\
			/* call the newstyle 9.12.0.1+ constructor*/ \
			plugin.plugin = new Class(com::softwareag::connectivity::PluginConstructorParameters::TransportConstructorParameters(name, chainId, com::softwareag::connectivity::get<com::softwareag::connectivity::map_t>(reinterpret_cast<const com::softwareag::connectivity::data_t&>(config)), connectivityManager, reserved));\
		} catch (...) {\
			com::softwareag::connectivity::handleException(com::softwareag::connectivity::Logger(std::string("connectivity.")+name+std::string(".")+chainId), SAG_LOG_ERROR, "create");\
		}\
		return plugin;\
	}\
}

////////////////////////////////////////////////////////////////////////////////////////////
// Transport chain managers

/** @internal */
#define _SAG_DECLARE_CONNECTIVITY_TRANSPORT_CHAIN_MANAGER_CLASS(Class) \
_SAG_DECLARE_CONNECTIVITY_TRANSPORT_COMMON(Class, Class::transport_t)\
extern "C" {\
void _sag_plugin_checkIsChainManager_##Class(Class *clazz) {\
	com::softwareag::connectivity::chainmanagers::AbstractChainManager<Class::transport_t> *cm = clazz; \
	(void) cm;\
}\
SAG_CONNECTIVITY_DLL_SYM sag_plugin_t SAG_CONNECTIVITY_CALL sag_chainmanager_create_##Class( sag_underlying_data_t config, const char *managerName, const char* transportPluginName, void *connectivityManager, sag_underlying_data_t chainDefs, void* reserved) {\
	sag_plugin_t plugin;\
	plugin.plugin = 0;\
	try {\
		plugin.plugin = new Class(com::softwareag::connectivity::chainmanagers::ChainManagerConstructorParameters(\
			com::softwareag::connectivity::get<com::softwareag::connectivity::map_t>(reinterpret_cast<const com::softwareag::connectivity::data_t&>(config)), \
			managerName, transportPluginName, connectivityManager, com::softwareag::connectivity::get<com::softwareag::connectivity::map_t>(reinterpret_cast<const com::softwareag::connectivity::data_t&>(chainDefs)), reserved));\
	} catch (...) {\
		com::softwareag::connectivity::handleException(com::softwareag::connectivity::Logger(std::string("connectivity.")+transportPluginName+"."+managerName), SAG_LOG_ERROR, "creating chain manager");\
	}\
	return plugin;\
}\
SAG_CONNECTIVITY_DLL_SYM sag_plugin_t SAG_CONNECTIVITY_CALL sag_chainmanager_createTransport_##Class(sag_plugin_t managerplugin, const char *chainId, sag_underlying_data_t config, void* connectivityManager, const char* defaultChannelTowardsHost, sag_underlying_vector_t subscribeChannels, void *fun, void* reserved) { \
	sag_plugin_t plugin;\
	plugin.plugin = 0;\
	Class *manager = reinterpret_cast<Class*>(managerplugin.plugin);\
	try {\
		auto &createTransport = *reinterpret_cast<std::function<Class::transport_t *(const com::softwareag::connectivity::chainmanagers::ManagedTransportConstructorParameters&)> *>(fun);\
		plugin.plugin = createTransport(com::softwareag::connectivity::chainmanagers::ManagedTransportConstructorParameters(manager->transportPluginName, chainId, \
			com::softwareag::connectivity::get<com::softwareag::connectivity::map_t>(reinterpret_cast<const com::softwareag::connectivity::data_t&>(config)), \
			connectivityManager, defaultChannelTowardsHost, \
			reinterpret_cast<const com::softwareag::connectivity::list_t&>(subscribeChannels).copy(),\
			reserved)); \
	} catch (...) {\
		com::softwareag::connectivity::handleException(manager->logger, SAG_LOG_ERROR, "createTransport");\
	}\
	return plugin;\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_chainmanager_start_##Class(sag_plugin_t plugin)\
{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->start();\
		return SAG_ERROR_OK;\
	} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::chainmanagers::AbstractChainManager<Class::transport_t>*>(plugin.plugin)->logger, SAG_LOG_ERROR, "starting chain manager");\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_chainmanager_shutdown_##Class(sag_plugin_t plugin)\
{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->shutdown();\
		return SAG_ERROR_OK;\
	} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::chainmanagers::AbstractChainManager<Class::transport_t>*>(plugin.plugin)->logger, SAG_LOG_WARN, "shutting down chain manager");\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_chainmanager_destroy_##Class(sag_plugin_t plugin)\
{\
	try {\
		delete reinterpret_cast<Class*>(plugin.plugin); \
		return SAG_ERROR_OK;\
	} catch (...) {\
		/*shouldn't throw, and we couldn't log it if it did*/\
		return SAG_ERROR_EXCEPTION;\
	}\
}\
}


////////////////////////////////////////////////////////////////////////////////////////////
// Codecs

/** @internal */
#define _SAG_DECLARE_CONNECTIVITY_CODEC_COMMON(Class) \
_SAG_DECLARE_CONNECTIVITY_CORE(Class, Class)  \
extern "C" {\
void _sag_plugin_checkIsCodec_##Class(Class *clazz) {\
	\
	com::softwareag::connectivity::AbstractCodec *ac = clazz;\
	(void) ac;\
}\
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_plugin_setNextTowardsTransport_##Class(sag_plugin_t plugin, sag_plugin_t next, sag_send_fn_t fn)\
	{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->setNextTowardsTransport(com::softwareag::connectivity::TransportSide::ptr_t(new com::softwareag::connectivity::RemoteTransportSide(next, fn)));\
		return SAG_ERROR_OK;\
			} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::Plugin*>(plugin.plugin)->logger, SAG_LOG_ERROR, "setNextTowardsTransport");\
		return SAG_ERROR_EXCEPTION;\
			}\
	} \
SAG_CONNECTIVITY_DLL_SYM sag_error_t SAG_CONNECTIVITY_CALL sag_plugin_sendBatchTowardsHost_##Class(sag_plugin_t plugin, sag_underlying_message_t *start, sag_underlying_message_t *end)\
	{\
	try {\
		reinterpret_cast<Class*>(plugin.plugin)->sendBatchTowardsHost(reinterpret_cast<com::softwareag::connectivity::Message *>(start), reinterpret_cast<com::softwareag::connectivity::Message *>(end));\
		return SAG_ERROR_OK;\
			} catch (...) {\
		com::softwareag::connectivity::handleException(reinterpret_cast<com::softwareag::connectivity::Plugin*>(plugin.plugin)->logger, SAG_LOG_WARN, "sendBatchTowardsHost");\
		return SAG_ERROR_EXCEPTION;\
			}\
	} \
}

/** @internal */
#define _SAG_DECLARE_CONNECTIVITY_CODEC_CLASS(Class) \
_SAG_DECLARE_CONNECTIVITY_CODEC_COMMON(Class)\
extern "C" {\
SAG_CONNECTIVITY_DLL_SYM sag_plugin_t SAG_CONNECTIVITY_CALL sag_create_plugin_with_params_##Class(const char *name, const char *chainId, sag_underlying_data_t config, void* correlatorInternal, void* reserved)\
	{\
	sag_plugin_t plugin;\
	plugin.plugin = 0;\
	try {\
		/* call the newstyle 9.12.0.1+ constructor*/ \
		plugin.plugin = new Class(com::softwareag::connectivity::PluginConstructorParameters::CodecConstructorParameters(name, chainId, com::softwareag::connectivity::get<com::softwareag::connectivity::map_t>(reinterpret_cast<const com::softwareag::connectivity::data_t&>(config)), correlatorInternal, reserved));\
	} catch (...) {\
	com::softwareag::connectivity::handleException(com::softwareag::connectivity::Logger(std::string("connectivity.")+name+std::string(".")+chainId), SAG_LOG_ERROR, "create");\
	}\
	return plugin;\
	}\
}


#endif // _SAG_INTERNAL_PLUGIN_MACROS_HPP_
